"""
==============================
PRIVATE CONFIGURATION SETTINGS
==============================

Global settings for:
*  ``kfunc`` deployment,
*  unit testing behaviour.
"""

# kfunc deployment (see .shortcuts)
KFUNC = 'shortcuts'

# testing (see .tests.shared)
PLOT = False
SAVE_ERRORS = False
VERBOSE = False
QUANT_TEST_MODE = 'LD'
QUANT_TEST_FAIL = True
